#INCLUDE "box.ch"
#INCLUDE "directry.ch"
#INCLUDE "common.ch"

static aLevel := {}, nLevel, nHandle

*!*****************************************************************************
*!
*!       Function: DUP_FIND()
*!
*!*****************************************************************************
function dup_find( cDrive, cTreeFile )

local cParentDir, nParent, nCurrentRec

field path, NAME, size, FDAT, FTIM, ATTR, PARENT

default cDrive to 'C:'
default cTreeFile to 'TREE.TXT'

cDrive := upper( cDrive )
cTreeFile := upper( cTreeFile )

setcolor( 'G+/B' )
clear screen

@ 3,29 say 'Duplicate File Finder'

dispbox(  5, 5, 13, 75, B_SINGLE + ' ' )
dispbox( 14, 5, 22, 75, B_SINGLE + ' ' )
setcolor( 'GR+/B' )

dbcreate('files.dbf',{ { 'PATH', 'C', 40, 0 },;
                       { 'NAME', 'C', 12, 0 },;
                       { 'SIZE', 'N', 10, 0 },;
                       { 'FDAT', 'D',  8, 0 },;
                       { 'FTIM', 'C',  8, 0 },;
                       { 'ATTR', 'C',  4, 0 } } )

dbcreate('tree.dbf',{ { 'PATH',   'C', 40, 0 },;
                      { 'COUNT',  'N', 10, 0 },;
                      { 'SIZE',   'N', 10, 0 },;
                      { 'PARENT', 'N',  8, 0 } } )

use files
// index on path to path
SHOW2( 'Scanning Directories on ' + cDrive )
DIR_RECURS( cDrive )

use TREE NEW
index on path to TREEPATH

select files
dbgotop()
do while !eof()
   select TREE
   if !dbseek( FILES->PATH )
      dbappend()
      TREE->PATH := FILES->PATH
      SHOW2( TREE->PATH )
   endif
   if !( 'D' $ FILES->ATTR )
      TREE->COUNT ++
      TREE->SIZE += FILES->SIZE
   endif
   select files
   skip
enddo

select TREE
go top
skip
do while !eof()
   cParentDir := left( path, rat( '\', path ) - 1 )
   nCurrentRec := recno()
   dbseek( cParentDir )
   nParent := recno()
   dbgoto( nCurrentRec )
   TREE->PARENT := nParent
   dbskip()
enddo

index on PARENT to PARENT

asize( aLevel, 30 )
afill( aLevel, .f. )
nLevel := 0
go top

nHandle := fcreate( cTreeFile, 0 )

Branch()

dupfind()

SHOW2( 'Tree and Duplicate text file is: ' + cTreeFile )

fclose( nHandle )

dbcloseall()

devpos( 22, 0 )

return NIL

*!*****************************************************************************
*!
*!      Procedure: DIR_RECURS
*!
*!*****************************************************************************
static function DIR_RECURS( cPath )

local x
local aFiles   := directory( cPath + '\*.*', 'D' )
local nFilCount := len( aFiles )

for x := 1 to nFilCount
   if aFiles[ X, F_NAME ] <> '..'
      dbappend()
      FILES->PATH := cPath
      FILES->NAME := aFiles[ X, F_NAME ]
      FILES->SIZE := aFiles[ X, F_SIZE ]
      FILES->FDAT := aFiles[ X, F_DATE ]
      FILES->FTIM := aFiles[ X, F_TIME ]
      FILES->ATTR := aFiles[ X, F_ATTR ]
      SHOW1( cPath + '\' + aFiles[ X, 1 ] )
   endif
   if 'D' $ aFiles[ X, F_ATTR ]
      if aFiles[ X, F_NAME ] <> '.'
         DIR_RECURS( cPath + '\' + aFiles[ X, F_NAME ] )
      endif
   endif
next

return NIL

*!*****************************************************************************
*!
*!       Function: BRANCH()
*!
*!*****************************************************************************
function Branch()

#define nIndent 2

local nParent, nCurrentRec, nNode, cThisLine

field PARENT, size, count, path

nLevel ++
cThisLine := transform( count, "9,999" ) + ' ' + transform( size, "999,999,999") + ' '
if nLevel > 1
   for nNode := 2 to nLevel - 1
      cThisLine += iif( aLevel[ nNode ], ' ', '' ) + space( nIndent )
   next
   nCurrentRec := recno()
   nParent := PARENT
   skip
   if nParent == PARENT
      cThisLine += '' + replicate( '', nIndent )
      aLevel[ nLevel ] = .f.
   else
      cThisLine += '' + replicate( '', nIndent )
      aLevel[ nLevel ] = .t.
   endif
   dbgoto( nCurrentRec )
endif
cThisLine += iif( len( alltrim( path ) ) = 2, path, substr( path, rat( '\', path ) + 1 ) )

SHOW2( cThisLine )

fwrite( nHandle, cThisLine + chr( 13 ) + chr( 10 ) )

nParent := recno()
dbseek( nParent )
do while PARENT == nParent
   nCurrentRec := recno()
   Branch()
   dbgoto( nCurrentRec )
   skip
enddo

nLevel --

return NIL

*!*****************************************************************************
*!
*!      Procedure: SHOW1
*!
*!*****************************************************************************
static function SHOW1( text )

scroll( 6, 6, 12, 74, 1 )

@ 12, 6 say text

return NIL

*!*****************************************************************************
*!
*!      Procedure: SHOW2
*!
*!*****************************************************************************
static function SHOW2( text )

scroll( 15, 6, 21, 74, 1 )

@ 21, 6 say text

return NIL

*!*****************************************************************************
*!
*!      Procedure: DUPFIND
*!
*!*****************************************************************************
function dupfind()

local crlf := chr( 13 ) + chr( 10 )
local lastpath, lastname, lastsize
local nNumDups := 0
local nDupBytes := 0

field path, name, size

show2( 'Creating Duplicate File index' )
select files
index on name + str( size, 10 ) to files

go top
do while files->( !eof() )
   lastpath := path
   lastname := name
   lastsize := size
   files->( dbskip() )
   if alltrim( name ) <> '.' .and. name == lastname .and. size == lastsize
      show1( left( 'Comparing ' + trim( lastpath ) + '\' + trim( lastname ) + ' to ' + trim( path ) + '\' + name, 67 ) )
      if filecomp( trim( lastpath ) + '\' + lastname, trim( path ) + '\' + name )
         fwrite( nHandle, space( 11 ) + trim( lastpath ) + '\' + lastname + CRLF )
         show2( trim( lastpath ) + '\' + lastname )
         fwrite( nHandle, str( size, 10 ) + ' ' + trim( path ) + '\' + name + CRLF + crlf )
         show2( trim( path ) + '\' + name )
         show2('')
         nNumDups ++
         nDupBytes += size
      else
         show1('Skipping ' + trim( lastpath ) + '\' + name )
      endif
   else
      show1('Skipping ' + trim( lastpath ) + '\' + name )
   endif
enddo

show2('')
show2( alltrim( str( nNumDups ) ) + ' duplicate files totalling ' + alltrim( str( nDupBytes ) ) + ' bytes' )
fwrite( nHandle, alltrim( str( nNumDups ) ) + ' duplicate files totalling ' + alltrim( str( nDupBytes ) ) + ' bytes' + crlf )

return NIL

*!*****************************************************************************
*!
*!      Procedure: FILECOMP
*!
*!*****************************************************************************
function filecomp( mainfile, testfile )

local retval     := .f.      // default is that files do not compare

#define BUFFLEN 4096

local buffer1    := space( BUFFLEN )
local bytesread1 := BUFFLEN
local handle1    := fopen( mainfile )
local buffer2    := space( BUFFLEN )
local bytesread2 := BUFFLEN
local handle2    := fopen( testfile )

if handle1 < 1 .or. handle2 < 1
   fclose( handle1 )
   fclose( handle2 )
   return retval
endif

retval := .t.            // now that they are open, default is true

do while bytesread1 == BUFFLEN

   bytesread1 := fread( handle1, @buffer1, BUFFLEN )

   if bytesread1 <> BUFFLEN
      buffer1 := left( buffer1, bytesread1 )
   endif

   bytesread2 := fread( handle2, @buffer2, BUFFLEN )

   if bytesread2 <> BUFFLEN
      buffer2 := left( buffer2, bytesread2 )
   endif

   if !( buffer1 == buffer2 )
      retval := .f.       // they failed synchronous comparison
      exit
   endif

enddo

fclose( handle1 )
fclose( handle2 )

return retval

*: EOF: DUP_FIND.PRG
